import json

from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import models
from django.db.models.fields import NOT_PROVIDED

from .forms import MatrixFormField
from .validators import DataTypeValidator, DimensionsValidator


class MatrixField(models.Field):
    """
    Provides database store for matrices.
    """

    description = "A matrix"
    __metaclass__ = models.SubfieldBase

    MAX_LENGTH = 255

    def __init__(self, datatype=None, dimensions=None, **kwargs):
        if not datatype or not dimensions:
            raise ImproperlyConfigured("Required kwarg 'datatype' or "
                                       "'dimensions' missing")
        self.datatype, self.dimensions = datatype, dimensions
        defaults = {
            'max_length': self.MAX_LENGTH,
            'default': None,
        }
        defaults.update(kwargs)
        super(MatrixField, self).__init__(**defaults)
        self.validators.append(DataTypeValidator(self.datatype))
        self.validators.append(DimensionsValidator(self.dimensions))

    def get_internal_type(self):
        # TODO: take advantage of postgres matrix type
        #       kinda the whole reason this django app even exists...
        return 'CharField'

    def to_python(self, value):
        "Convert to python matrix (arrays of arrays)"
        if isinstance(value, (list, tuple)) or value is None:
            return value
        try:
            return json.loads(value)
        except:
            raise ValidationError("Unable to convert value to matrix")

    def get_prep_value(self, value):
        "Convert to string"
        return json.dumps(value)

    def formfield(self, **kwargs):
        defaults = {
            'datatype': self.datatype,
            'dimensions': self.dimensions,
            'form_class': MatrixFormField,
        }
        defaults.update(kwargs)
        return super(MatrixField, self).formfield(**defaults)


# South support
try:
    from south.modelsinspector import add_introspection_rules
except ImportError:
    pass
else:
    add_introspection_rules(
        rules=[(
            (MatrixField, ),    # Class(es) these apply to
            [],                 # Positional arguments (not used)
            {                   # Keyword argument
                'max_length': [
                    'max_length', {'default': MatrixField.MAX_LENGTH}
                ],
                'datatype': ['datatype', {'default': NOT_PROVIDED}],
                'dimensions': ['dimensions', {'default': NOT_PROVIDED}],
            },
        )],
        patterns=['matrix_field\.fields\.']
    )



import json
from django import forms
from .validators import DataTypeValidator, DimensionsValidator


class MatrixFormField(forms.Field):

    def __init__(self, datatype=None, dimensions=None, **kwargs):
        super(MatrixFormField, self).__init__(**kwargs)
        self.validators.append(DataTypeValidator(datatype=datatype))
        self.validators.append(DimensionsValidator(dimensions=dimensions))

    def to_python(self, value):
        if value is None or isinstance(value, (list, tuple)):
            return value
        return json.loads(value)



from django.core.exceptions import ValidationError


class DataTypeValidator(object):
    "Validate that each element of a matrix is of the correct type"

    def __init__(self, datatype):
        super(DataTypeValidator, self).__init__()
        # str and unicode are mapped to 'database store for strings'
        if datatype in ('list', 'tuple'):
            raise ValidationError("Invalid datatype - not a primitive type")
        self.datatype = (
            'basestring' if datatype in ('str', 'unicode')
            else datatype
        )

    def __call__(self, value):
        if isinstance(value, (list, tuple)):
            for elem in value:
                self(elem)
        else:
            datatype = type(value).__name__
            if datatype in ('str', 'unicode'):
                datatype = 'basestring'
            if datatype != self.datatype:
                raise ValidationError("Invalid datatype")


class DimensionsValidator(object):
    "Validate the matrix has correct dimensions"

    def __init__(self, dimensions):
        super(DimensionsValidator, self).__init__()
        self.dimensions = dimensions

    def __call__(self, value, index=0):
        if index == len(self.dimensions):
            if isinstance(value, (list, tuple)):
                raise ValidationError("Excessive depth of dimensions")
            return
        if not isinstance(value, (list, tuple)):
            raise ValidationError("Insufficient depth of dimensions")
        if not len(value) == self.dimensions[index]:
            raise ValidationError("Invalid dimension")
        for elem in value:
            self(elem, index+1)



import json

from django import forms
from django.core.exceptions import ValidationError
from django.db import models
from django.test import TestCase

from . import MatrixField, MatrixFormField


M1 = [[3, 4], [4, 5], [6, 7]]
M2 = [[[2.0]]]
M3a = ['a', 'b']
M3b = ['c', 'd']

M1_str = json.dumps(M1)
M2_str = json.dumps(M2)
M3a_str = json.dumps(M3a)
M3b_str = json.dumps(M3b)


class TestModel(models.Model):
    f1 = MatrixField(datatype='int', dimensions=(3, 2))
    f2 = MatrixField(datatype='float', dimensions=(1, 1, 1), blank=True)
    f3 = MatrixField(datatype='str', dimensions=(2,), blank=True,
                     default=M3a_str)


class TestForm(forms.Form):
    f1 = MatrixFormField(datatype='int', dimensions=(3, 2))
    f2 = MatrixFormField(datatype='float', dimensions=(1, 1, 1),
                         required=False)


class TestModelForm(forms.ModelForm):
    class Meta:
        model = TestModel


class MatrixFormFieldTestCase(TestCase):

    def test_valid_specify_all(self):
        form = TestForm({'f1': M1_str, 'f2': M2_str})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['f1'], M1)
        self.assertEqual(form.cleaned_data['f2'], M2)

    def test_valid_with_defaults(self):
        form = TestForm({'f1': M1_str})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['f1'], M1)
        self.assertEqual(form.cleaned_data['f2'], None)

    def test_invalid_blank(self):
        form = TestForm({})
        self.assertFalse(form.is_valid())
        self.assertTrue(any('required' in e for e in form.errors['f1']))

    def test_invalid_datatype(self):
        value = [[1.0, 1.0], [1.0, 1.0], [1.0, 1.0]]
        form = TestForm({'f1': json.dumps(value)})
        self.assertFalse(form.is_valid())
        self.assertTrue(any('datatype' in e for e in form.errors['f1']))

    def test_invalid_dimesion(self):
        form = TestForm({'f1': json.dumps([1])})
        self.assertFalse(form.is_valid())
        self.assertTrue(any('dimension' in e for e in form.errors['f1']))


class MatrixFieldModelFormTestCase(TestCase):

    def test_valid_specify_all(self):
        form = TestModelForm({'f1': M1_str, 'f2': M2_str, 'f3': M3b_str})
        self.assertTrue(form.is_valid())
        form.save()
        self.assertEqual(TestModel.objects.count(), 1)

        m = TestModel.objects.get()
        self.assertEqual(m.f1, M1)
        self.assertEqual(m.f2, M2)
        self.assertEqual(m.f3, M3b)

    def test_valid_with_defaults(self):
        # seems there should be a better way to get a form's default values...?
        # http://stackoverflow.com/questions/7399490/
        data = dict(
            (field_name, field.initial)
            for field_name, field in TestModelForm().fields.iteritems()
        )
        data.update({'f1': M1})
        form = TestModelForm(data=data)
        self.assertTrue(form.is_valid())
        form.save()
        self.assertEqual(TestModel.objects.count(), 1)

        m = TestModel.objects.get()
        self.assertEqual(m.f2, None)
        self.assertEqual(m.f3, M3a)

    def test_invalid_required(self):
        form = TestModelForm({})
        self.assertFalse(form.is_valid())
        self.assertTrue(any('required' in e for e in form.errors['f1']))

    def test_invalid_datatype(self):
        value = [[1.0, 1.0], [1.0, 1.0], [1.0, 1.0]]
        form = TestModelForm({'f1': json.dumps(value)})
        self.assertFalse(form.is_valid())
        self.assertTrue(any('datatype' in e for e in form.errors['f1']))

    def test_invalid_dims1(self):
        form = TestModelForm({'f1': M1_str, 'f2': json.dumps([[1]])})
        self.assertFalse(form.is_valid())
        self.assertTrue(any('dimension' in e for e in form.errors['f2']))

    def test_invalid_dims2(self):
        value = ['a', 'b', 'c']
        form = TestModelForm({'f1': M1_str, 'f3': json.dumps(value)})
        self.assertFalse(form.is_valid())
        self.assertTrue(any('dimension' in e for e in form.errors['f3']))


class MatrixFieldTestCase(TestCase):

    def test_valid(self):
        m = TestModel.objects.create(f1=M1, f2=M2, f3=M3b)
        m = TestModel.objects.get(pk=m.pk)
        self.assertEqual(m.f1, M1)
        self.assertEqual(m.f2, M2)
        self.assertEqual(m.f3, M3b)

    def test_valid_blank(self):
        m = TestModel.objects.create(f1=M1)
        m = TestModel.objects.get(pk=m.pk)
        self.assertEqual(m.f2, None)
        self.assertEqual(m.f3, M3a)

    def test_invalid_type(self):
        with self.assertRaises(ValidationError):
            TestModel(f1=M1, f3=[2.0, 3.0]).full_clean()
        with self.assertRaises(ValidationError):
            TestModel(f1=M1, f2=[[[2]]]).full_clean()

    def test_invalid_dims(self):
        with self.assertRaises(ValidationError):
            TestModel(f1=M1, f2=['a']).full_clean()
        with self.assertRaises(ValidationError):
            TestModel(f1=M1, f3=[['a'], ['b']]).full_clean()

    def test_invalid_altogether(self):
        with self.assertRaises(ValidationError):
            TestModel(f1={'a': 2}).full_clean()
        with self.assertRaises(ValidationError):
            TestModel(f1=4).full_clean()
        with self.assertRaises(ValidationError):
            TestModel(f1='blah blah').full_clean()
